/* $Id: we_plugins.c,v 1.10 2007/11/25 18:19:35 ekalin Exp $ */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 * MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <kcconfig.h>
#endif

#include <string.h>
#include <libintl.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>

#include "kildclient.h"
#include "perlscript.h"


/*****************************************
 * Function prototypes - Startup Plugins *
 *****************************************/
static void     configure_startup_view(GtkTreeView *view, World *world);
static void     we_splugin_add_cb(GtkButton *button, gpointer data);
static void     we_splugin_remove_cb(GtkButton *button, gpointer data);
static gboolean we_splugin_view_keypress_cb(GtkWidget   *widget,
                                            GdkEventKey *evt,
                                            gpointer     data);
static void     we_splugin_move_cb(GtkButton *button, gpointer data);

/**************************************************
 * Function prototypes - Currently Loaded Plugins *
 **************************************************/
static GtkTreeModel *create_we_lplugins_model(World *world);
static void          configure_loaded_view(GtkTreeView  *view,
                                           GtkTreeModel *model);
static void          we_lplugin_ena_toggled_cb(GtkCellRendererToggle *renderer,
                                               gchar                 *pathstr,
                                               gpointer               data);
static void          we_lplugin_load_cb(GtkButton *button, gpointer data);
static void          we_lplugin_help_cb(GtkButton *button, gpointer data);


/*********************************
 * Function prototypes - Generic *
 *********************************/
static GtkWidget *we_plugin_get_load_filechooser(GtkWindow *parent);



GtkWidget *
get_we_plugins_tab(GladeXML *gladexml, World *world)
{
  GtkWidget *panel;

  GtkWidget *lstStartupPlugins;
  GtkWidget *btnSPluginAdd;
  GtkWidget *btnSPluginRemove;
  GtkWidget *btnSPluginUp;
  GtkWidget *btnSPluginDown;

  GtkWidget *lstLoadedPlugins;
  GtkWidget *btnLPluginLoad;
  GtkWidget *btnLPluginHelp;

  panel = glade_xml_get_widget(gladexml, "we_panel_plugins");

  /* Currently loaded plugins part */
  lstLoadedPlugins = glade_xml_get_widget(gladexml, "lstLoadedPlugins");
  g_object_set_data(G_OBJECT(lstLoadedPlugins), "forworld", world);

  btnLPluginLoad = glade_xml_get_widget(gladexml, "btnLPluginLoad");
  g_signal_connect(G_OBJECT(btnLPluginLoad), "clicked",
                   G_CALLBACK(we_lplugin_load_cb), lstLoadedPlugins);
  if (!world->connected) {
    gtk_widget_set_sensitive(btnLPluginLoad, FALSE);
  }

  btnLPluginHelp = glade_xml_get_widget(gladexml, "btnLPluginHelp");
  g_signal_connect(G_OBJECT(btnLPluginHelp), "clicked",
                   G_CALLBACK(we_lplugin_help_cb), lstLoadedPlugins);

  world->plugin_model = create_we_lplugins_model(world);
  configure_loaded_view(GTK_TREE_VIEW(lstLoadedPlugins), world->plugin_model);


  /* Startup plugins part */
  lstStartupPlugins = glade_xml_get_widget(gladexml, "lstStartupPlugins");
  g_object_set_data(G_OBJECT(lstStartupPlugins), "forworld", world);
  g_signal_connect(G_OBJECT(lstStartupPlugins), "key-press-event",
                   G_CALLBACK(we_splugin_view_keypress_cb), world);

  btnSPluginAdd = glade_xml_get_widget(gladexml, "btnSPluginAdd");
  g_signal_connect(G_OBJECT(btnSPluginAdd), "clicked",
                   G_CALLBACK(we_splugin_add_cb), lstStartupPlugins);
  btnSPluginRemove = glade_xml_get_widget(gladexml, "btnSPluginRemove");
  g_signal_connect(G_OBJECT(btnSPluginRemove), "clicked",
                   G_CALLBACK(we_splugin_remove_cb), lstStartupPlugins);

  btnSPluginUp = glade_xml_get_widget(gladexml, "btnSPluginUp");
  g_object_set_data(G_OBJECT(btnSPluginUp), "isup", GINT_TO_POINTER(TRUE));
  g_signal_connect(G_OBJECT(btnSPluginUp), "clicked",
                   G_CALLBACK(we_splugin_move_cb), lstStartupPlugins);
  btnSPluginDown = glade_xml_get_widget(gladexml, "btnSPluginDown");
  g_object_set_data(G_OBJECT(btnSPluginDown), "isup", GINT_TO_POINTER(FALSE));
  g_signal_connect(G_OBJECT(btnSPluginDown), "clicked",
                   G_CALLBACK(we_splugin_move_cb), lstStartupPlugins);

  configure_startup_view(GTK_TREE_VIEW(lstStartupPlugins), world);

  return panel;
}


/* Startup plugins */
static
void
configure_startup_view(GtkTreeView *view, World *world)
{
  GtkCellRenderer   *renderer;
  GtkTreeViewColumn *column;

  gtk_tree_view_set_model(view, world->startup_plugins);

  if (world->connected) {
    /* Column 1 - Name */
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_column_set_expand(column, TRUE);
    gtk_tree_view_column_set_title(column, _("Name"));
    gtk_tree_view_column_pack_start(column, renderer, TRUE);
    gtk_tree_view_column_add_attribute(column,
                                       renderer,
                                       "text", SPLUGIN_NAME);
    gtk_tree_view_append_column(view, column);

    /* Column 2 - Description */
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_column_set_expand(column, TRUE);
    gtk_tree_view_column_set_title(column, _("Description"));
    gtk_tree_view_column_pack_start(column, renderer, TRUE);
    gtk_tree_view_column_add_attribute(column,
                                       renderer,
                                       "text", SPLUGIN_DESCRIPTION);
    gtk_tree_view_append_column(view, column);

    /* Column 3 - Author */
    renderer = gtk_cell_renderer_text_new();
    column = gtk_tree_view_column_new();
    gtk_tree_view_column_set_resizable(column, TRUE);
    gtk_tree_view_column_set_expand(column, TRUE);
    gtk_tree_view_column_set_title(column, _("Author"));
    gtk_tree_view_column_pack_start(column, renderer, TRUE);
    gtk_tree_view_column_add_attribute(column,
                                       renderer,
                                       "text", SPLUGIN_AUTHOR);
    gtk_tree_view_append_column(view, column);
  }

  /* Column 4 - File */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_column_set_title(column, _("File"));
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_add_attribute(column,
                                     renderer,
                                     "text", SPLUGIN_FILE);
  gtk_tree_view_append_column(view, column);


  gtk_tree_view_columns_autosize(view);
}


static
void
we_splugin_add_cb(GtkButton *button, gpointer data)
{
  World       *world;
  GtkTreeView *view;
  GtkWidget   *filechooser;

  view = (GtkTreeView *) data;
  world = (World *) g_object_get_data(G_OBJECT(view), "forworld");

  filechooser = we_plugin_get_load_filechooser(GTK_WINDOW(world->dlgEditWorld));

  if (gtk_dialog_run(GTK_DIALOG(filechooser)) == GTK_RESPONSE_YES) {
    gchar  *filename;
    GError *error = NULL;

    gtk_widget_hide(filechooser);
    filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(filechooser));
    if (world->connected) {
      add_startup_plugin(world, filename, &error);
    } else {
      GtkTreeIter iter;

      gtk_list_store_append(GTK_LIST_STORE(world->startup_plugins), &iter);
      gtk_list_store_set(GTK_LIST_STORE(world->startup_plugins), &iter,
                         SPLUGIN_FILE, filename,
                         -1);

    }
    g_free(filename);
    if (error) {
      GtkWidget *dlg = gtk_message_dialog_new(GTK_WINDOW(world->dlgEditWorld),
                                              GTK_DIALOG_MODAL,
                                              GTK_MESSAGE_ERROR,
                                              GTK_BUTTONS_OK,
                                              error->message);
      gtk_window_set_title(GTK_WINDOW(dlg), _("Error Loading Plugin"));
      gtk_dialog_run(GTK_DIALOG(dlg));
      gtk_widget_destroy(dlg);
      g_error_free(error);
    }
  }
  gtk_widget_destroy(filechooser);
}


static
void
we_splugin_remove_cb(GtkButton *button, gpointer data)
{
  GtkTreeView      *view;
  GtkTreeModel     *model;
  GtkTreeSelection *selection;
  GtkTreeIter       iter;

  view = (GtkTreeView *) data;
  selection = gtk_tree_view_get_selection(view);
  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
  }
}


static
void
we_splugin_move_cb(GtkButton *button, gpointer data)
{
  GtkTreeView      *view;
  GtkTreeModel     *model;
  GtkTreeSelection *selection;
  GtkTreeIter       iter;
  GtkTreeIter       other_iter;
  gboolean          is_up;

  view = (GtkTreeView *) data;
  selection = gtk_tree_view_get_selection(view);
  is_up = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(button), "isup"));
  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    if (is_up) {
      GtkTreePath *path;

      path = gtk_tree_model_get_path(model, &iter);
      if (gtk_tree_path_prev(path)) {
        gtk_tree_model_get_iter(model, &other_iter, path);
        gtk_list_store_swap(GTK_LIST_STORE(model), &iter, &other_iter);
      }
      gtk_tree_path_free(path);
    } else {
      other_iter = iter;
      if (gtk_tree_model_iter_next(model, &other_iter)) {
        gtk_list_store_swap(GTK_LIST_STORE(model), &iter, &other_iter);
      }
    }
  }
}


static
gboolean
we_splugin_view_keypress_cb(GtkWidget   *widget,
                            GdkEventKey *evt,
                            gpointer     data)
{
  if (evt->keyval == GDK_Delete || evt->keyval == GDK_KP_Delete) {
    we_splugin_remove_cb(NULL, widget);
    return TRUE;
  }

  return FALSE;
}


/* Currently loaded plugins */
static
GtkTreeModel *
create_we_lplugins_model(World *world)
{
  GtkListStore *store;
  GtkTreeIter   iter;
  GSList       *pluginptr;
  Plugin       *plugin;

  store = gtk_list_store_new(WE_GUIED_N_COLS,
                             G_TYPE_POINTER);

  pluginptr = world->plugins;
  while (pluginptr) {
    plugin = (Plugin *) pluginptr->data;
    gtk_list_store_append(store, &iter);
    gtk_list_store_set(store, &iter,
                       WE_GUIED_POINTER, plugin,
                       -1);

    pluginptr = pluginptr->next;
  }

  return GTK_TREE_MODEL(store);
}


static
void
configure_loaded_view(GtkTreeView *view, GtkTreeModel *model)
{
  GtkCellRenderer   *renderer;
  GtkTreeViewColumn *column;

  gtk_tree_view_set_model(view, model);
  g_object_unref(model);

  /* Column 1 - Enabled? */
  renderer = gtk_cell_renderer_toggle_new();
  g_object_set(G_OBJECT(renderer), "activatable", TRUE, NULL);
  g_signal_connect(G_OBJECT(renderer), "toggled",
                   G_CALLBACK(we_lplugin_ena_toggled_cb), view);
  column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_title(column, _("Enabled"));
  gtk_tree_view_column_pack_start(column, renderer, FALSE);
  gtk_tree_view_column_set_cell_data_func(column,
                                          renderer,
                                          we_guied_bool_func,
                                          GINT_TO_POINTER(PLUGIN_ENABLED),
                                          NULL);
  gtk_tree_view_append_column(view, column);

  /* Column 2 - Name */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_column_set_title(column, _("Name"));
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_set_cell_data_func(column,
                                          renderer,
                                          we_guied_text_func,
                                          GINT_TO_POINTER(PLUGIN_NAME),
                                          NULL);
  gtk_tree_view_append_column(view, column);

  /* Column 3 - Version */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_column_set_title(column, _("Version"));
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_set_cell_data_func(column,
                                          renderer,
                                          we_guied_text_func,
                                          GINT_TO_POINTER(PLUGIN_VERSION),
                                          NULL);
  gtk_tree_view_append_column(view, column);

  /* Column 4 - Author */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_column_set_title(column, _("Author"));
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_set_cell_data_func(column,
                                          renderer,
                                          we_guied_text_func,
                                          GINT_TO_POINTER(PLUGIN_AUTHOR),
                                          NULL);
  gtk_tree_view_append_column(view, column);

  /* Column 5 - Description */
  renderer = gtk_cell_renderer_text_new();
  column = gtk_tree_view_column_new();
  gtk_tree_view_column_set_resizable(column, TRUE);
  gtk_tree_view_column_set_expand(column, TRUE);
  gtk_tree_view_column_set_title(column, _("Description"));
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_set_cell_data_func(column,
                                          renderer,
                                          we_guied_text_func,
                                          GINT_TO_POINTER(PLUGIN_DESCRIPTION),
                                          NULL);
  gtk_tree_view_append_column(view, column);


  gtk_tree_view_columns_autosize(view);
}

static
void
we_lplugin_ena_toggled_cb(GtkCellRendererToggle *renderer,
                          gchar                 *pathstr,
                          gpointer               data)
{
  GtkTreeView      *view;
  GtkTreeModel     *model;
  GtkTreeSelection *selection;
  GtkTreeIter       iter;
  World            *world;
  Plugin           *plugin;

  view      = (GtkTreeView *) data;
  model     = gtk_tree_view_get_model(view);
  selection = gtk_tree_view_get_selection(view);
  world     = (World *) g_object_get_data(G_OBJECT(view), "forworld");
  gtk_tree_model_get_iter_from_string(model, &iter, pathstr);
  if (gtk_tree_selection_iter_is_selected(selection, &iter)) {
    gchar *cmd;

    gtk_tree_model_get(model, &iter, WE_GUIED_POINTER, &plugin, -1);
    if (plugin->enabled) {
      cmd = g_strdup_printf("$::world->displugin('%s');", plugin->name);
    } else {
      cmd = g_strdup_printf("$::world->enaplugin('%s');", plugin->name);
    }
    PERL_SET_CONTEXT(world->perl_interpreter);
    eval_pv(cmd, TRUE);
    g_free(cmd);
  }
}


static
void
we_lplugin_load_cb(GtkButton *button, gpointer data)
{
  World        *world;
  GtkTreeView  *view;
  GtkWidget    *filechooser;
  GtkWidget    *chkAtStartup;

  view = (GtkTreeView *) data;
  world = (World *) g_object_get_data(G_OBJECT(view), "forworld");

  filechooser = we_plugin_get_load_filechooser(GTK_WINDOW(world->dlgEditWorld));

  chkAtStartup = gtk_check_button_new_with_mnemonic(_("Always load this plugin at _startup"));
  gtk_file_chooser_set_extra_widget(GTK_FILE_CHOOSER(filechooser),
                                    chkAtStartup);

  if (gtk_dialog_run(GTK_DIALOG(filechooser)) == GTK_RESPONSE_YES) {
    gchar  *filename;
    GError *error = NULL;

    gtk_widget_hide(filechooser);
    filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(filechooser));
    if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(chkAtStartup))) {
      add_startup_plugin(world, filename, &error);
    }
    if (error == NULL) {
      load_plugin(world, filename, &error);
    }
    if (error) {
      GtkWidget *dlg = gtk_message_dialog_new(GTK_WINDOW(world->dlgEditWorld),
                                              GTK_DIALOG_MODAL,
                                              GTK_MESSAGE_ERROR,
                                              GTK_BUTTONS_OK,
                                              error->message);
      gtk_window_set_title(GTK_WINDOW(dlg), _("Error Loading Plugin"));
      gtk_dialog_run(GTK_DIALOG(dlg));
      gtk_widget_destroy(dlg);
      g_error_free(error);
    }
  }
  gtk_widget_destroy(filechooser);
}


static
void
we_lplugin_help_cb(GtkButton *button, gpointer data)
{
  World            *world;
  GtkTreeView      *view;
  GtkTreeModel     *model;
  GtkTreeSelection *selection;
  GtkTreeIter       iter;

  view = (GtkTreeView *) data;
  world = (World *) g_object_get_data(G_OBJECT(view), "forworld");
  selection = gtk_tree_view_get_selection(view);
  if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
    Plugin    *plugin;
    gchar     *cmd;
    GtkWidget *msgdlg;

    gtk_tree_model_get(model, &iter, WE_GUIED_POINTER, &plugin, -1);
    cmd = g_strdup_printf("%s::help();", plugin->name);
    PERL_SET_CONTEXT(world->perl_interpreter);
    eval_pv(cmd, FALSE);
    g_free(cmd);

    /* Show message telling where the help appears. */
    if (!globalPrefs.no_plugin_help_msg) {
      GtkWidget *chk;
      msgdlg = gtk_message_dialog_new(GTK_WINDOW(world->dlgEditWorld),
                                      GTK_DIALOG_MODAL,
                                      GTK_MESSAGE_INFO,
                                      GTK_BUTTONS_OK,
                                      _("The help is displayed in main window."));
      gtk_window_set_title(GTK_WINDOW(msgdlg), _("KildClient Information"));
      chk = gtk_check_button_new_with_mnemonic(_("_Do not display this message again"));
      gtk_box_pack_start(GTK_BOX(GTK_DIALOG(msgdlg)->vbox), chk,
                         FALSE, FALSE, 4);

      gtk_widget_show_all(msgdlg);
      gtk_dialog_run(GTK_DIALOG(msgdlg));

      if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(chk))) {
        globalPrefs.no_plugin_help_msg = TRUE;
      }

      gtk_widget_destroy(msgdlg);
    }
  }
}


void
we_lplugin_insert_plugin(World *world, Plugin *plugin)
{
  /* Called by the Perl functions when an plugin is inserted, so that
     the display is updated. */
  GtkTreeIter iter;

  if (!world->plugin_model) {
    return;
  }

  gtk_list_store_append(GTK_LIST_STORE(world->plugin_model), &iter);
  gtk_list_store_set(GTK_LIST_STORE(world->plugin_model), &iter,
                     WE_GUIED_POINTER, plugin,
                     -1);
}


void
we_lplugin_update_plugin(World *world, Plugin *plugin_arg)
{
  /* Called by the Perl functions when an plugin is changed, so that
     the display is updated. */
  GtkTreeIter  iter;
  GtkTreePath *path;
  Plugin      *plugin;
  gboolean     success;

  if (!world->plugin_model) {
    return;
  }

  success = gtk_tree_model_get_iter_first(world->plugin_model, &iter);
  while (success) {
    gtk_tree_model_get(world->plugin_model, &iter,
                       WE_GUIED_POINTER, &plugin, -1);
    if (plugin == plugin_arg) {
      path = gtk_tree_model_get_path(world->plugin_model, &iter);
      gtk_tree_model_row_changed(world->plugin_model, path, &iter);
      gtk_tree_path_free(path);
      return;
    }
    success = gtk_tree_model_iter_next(world->plugin_model, &iter);
  }
}


static
GtkWidget *
we_plugin_get_load_filechooser(GtkWindow *parent)
{
  const gchar *plugindir;
  GtkWidget   *filechooser;

  filechooser = gtk_file_chooser_dialog_new(_("Load plugin"),
                                           parent,
                                           GTK_FILE_CHOOSER_ACTION_OPEN,
                                           GTK_STOCK_CANCEL,
                                           GTK_RESPONSE_CANCEL,
                                           GTK_STOCK_OPEN,
                                           GTK_RESPONSE_YES,
                                           NULL);
  plugindir = get_kildclient_installed_file("plugins");
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(filechooser),
                                      plugindir);
  gtk_file_chooser_add_shortcut_folder(GTK_FILE_CHOOSER(filechooser),
                                       plugindir,
                                       NULL);
#ifndef __WIN32__
  plugindir = g_strdup_printf("%s/plugins",
                              get_kildclient_directory_path());
#else
  plugindir = g_strdup_printf("%s\\plugins",
                              get_kildclient_directory_path());
#endif
  gtk_file_chooser_add_shortcut_folder(GTK_FILE_CHOOSER(filechooser),
                                       plugindir,
                                       NULL);

  return filechooser;
}
